; This file is part of PALESTRA
;
; PALESTRA v0.1, abr-2021, GPL-2.0-or-later, vlan7 <https://www.vlan7.org>
; un TSR para traducir al castellano la aventura grafica/conversacional
; Les Manley in: Search for the King v2.0
;
; PALESTRA is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 2 of the License, or
; (at your option) any later version.
;
; PALESTRA is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with PALESTRA.  If not, see <https://www.gnu.org/licenses/>.
;
; why PALESTRA?
; $ grep les /usr/share/dict/spanish |grep tra
; palestra

; ****************************************
; *                                      *
; *  R U T I N A S     S T R I N G . H   *
; *                                      *
; ****************************************

;strlen: calcula longitud cadena terminada en \x00
;strcmp: compara dos cadenas
;strcpy: copia una cadena en otra cadena, terminando esta ultima con \0
;strstr: buscamos una subcadena en una cadena
;indicecadena:

;------------------------------------------------------------------------
; STRLEN: Calcula longitud cadena terminada en \x00
; 
; IN:
;     push offset cadena
;     call strlen
; 
; OUT: 
;     ax=longitud cadena
; 
; OJO!! 
;   1. Asume que la cadena esta en DS
;   2. Asume cadena terminada en \x00
;------------------------------------------------------------------------
strlen proc near
    push bp
    mov bp,sp
    push    cx,si           ;backup

;ahora la pila es
;bp+0 => old BP
;bp+2 => @ret
;bp+4 => offset cadena
;registros pusheados
 
    mov si,word [bp+04h]    ;cadena
    mov cx,si

@@countchars:   lodsw       ;MOV word en DS:SI a AX
    and ax,7f7fh            ;mascara: bit mas alto de cada byte
    sub ax,0101h
    test ax,8080h
    jz @@countchars
    mov ax,Word Ptr [si-2]
    test al,al              ;¿quedan caracteres por recorrer?
    je @@fin_de_cadena
    inc si                  ;si, siguiente caracter
    test ah,ah              ;¿quedan caracteres por recorrer?
    je @@fin_de_cadena
    inc si                  ;si, siguiente caracter
@@fin_de_cadena:    lea ax,[si-2]
    sub ax,cx               ;ax=longitud
    pop si,cx               ;restore
    mov sp,bp
    pop bp
    ret 2                   ;2*(1 parametro)
strlen endp

;------------------------------------------------------------------------
;fast_strcmp: Compara cadena en ds:[si] con cadena en es:[di]
;in: nada
;out: ax=0, match
;     ax=-1, no match
;asume ¿cadenas de al menos 1 caracter? y terminadas en \x00
;------------------------------------------------------------------------
fast_strcmp proc near
    push    si,di                   ;backup
@@loop_fast_strcmp: cmpsb
    jne @@no_match_fast_strcmp
    cmp Byte Ptr ds:[si-1],0                               ;match, por ahora
jnz @@loop_fast_strcmp
    xor ax,ax                       ;match
    pop di,si       ;restore
    ret
@@no_match_fast_strcmp: or ax,-1    ;no match
    pop di,si       ;restore
    ret
fast_strcmp endp

;------------------------------------------------------------------------
;faster_strcmp: Compara cadena en ds:[si] con cadena en es:[di]
;destroy: si,di
;in: nada
;out: ax=0, match
;     ax=-1, no match
;asume ¿cadenas de al menos 1 caracter? y terminadas en \x00
;------------------------------------------------------------------------
faster_strcmp proc near
@@loop_faster_strcmp:   cmpsb
    jne @@no_match_faster_strcmp
    cmp Byte Ptr ds:[si-1],0                               ;match, por ahora
jnz @@loop_faster_strcmp
    xor ax,ax                           ;match
    ret
@@no_match_faster_strcmp:   or ax,-1    ;no match
    ret
faster_strcmp endp

;------------------------------------------------------------------------
; strcmp: Compara cadena s1 con cadena s2
; 
; IN:
;     push segmento cadena destino (s1)
;     push offset cadena destino (s1)
;     push segmento cadena origen (s2)
;     push offset cadena origen (s2)
;     call STRCMP
; 
; OUT: 
;     AX - 0 si son iguales, -1 si son diferentes
; 
; OJO!!
;   1. Asume cadenas terminadas en \x00
;------------------------------------------------------------------------
strcmp  proc    near
    push bp
    mov bp,sp
    push si,di,es,ds        ;backup

; ahora la pila es
; bp+0 => old BP
; bp+2 => @ret
; bp+4 => offset s2 (origen)
; bp+6 => segmento s2 (origen)
; bp+8 => offset s1 (destino)
; bp+0a => segmento s1 (destino)
; registros pusheados

    _src_offset      equ [word bp+4]
    _src_segmento    equ [word bp+6]
    _dst_offset      equ [word bp+8]
    _dst_segmento    equ [word bp+0a]

    mov es,_src_segmento
    mov ds,_dst_segmento
    mov di,_src_offset
    mov si,_dst_offset

@@loop_strcmp:  lodsb
    cmp al,byte es:[di]
    jne @@strcmp_nomatch
    test al,al
    jz @@strcmp_match
    inc di
jmp short @@loop_strcmp
@@strcmp_match: xor ax,ax
    jmp short @@exit_strcmp
@@strcmp_nomatch:   mov ax,-1
@@exit_strcmp:  pop ds,es,di,si     ;restore
    mov sp,bp
    pop bp
    ret 8                           ;2*(4 params)
strcmp  endp

;------------------------------------------------------------------------
; strcpy: Copia cadena s2 en cadena s1, terminando s1 con \0
; 
; IN:
;     push segmento cadena destino (s1)
;     push offset cadena destino (s1)
;     push segmento cadena origen (s2)
;     push offset cadena origen (s2)
;     call STRCPY
; 
; OJO!!
;   1. Asume cadenas terminadas en \x00
;   2. Asume que s1 (destino) es suficientemente grande
;------------------------------------------------------------------------
strcpy  proc    near
    push bp
    mov bp,sp
    push cx,si,di,es,ds ;backup
 
; ahora la pila es
; bp+0 => old BP
; bp+2 => @ret
; bp+4 => offset s2 (origen)
; bp+6 => segmento s2 (origen)
; bp+8 => offset s1 (destino)
; bp+0a => segmento s1 (destino)
; registros pusheados

    _src_offset      equ [word bp+4]
    _src_segmento    equ [word bp+6]
    _dst_offset      equ [word bp+8]
    _dst_segmento    equ [word bp+0a]
 
    mov ds,_src_segmento
    push _src_offset
    call strlen
    test ax,ax
    je @@end_strcpy
    mov cx,ax                   ;¿optimizar con xchg cx,ax?
    mov es,_dst_segmento
    mov di,_dst_offset
    mov si,_src_offset
    rep movsb                   ;MOV CX bytes de DS:SI a ES:DI
;    mov word es:[di],0020h     ;espacio + fin de cadena
    mov byte es:[di],0          ;fin de cadena
@@end_strcpy:   pop ds,es,di,si,cx  ;restore
    mov sp,bp
    pop bp
    ret 8                       ;2*(4 parametros)
strcpy  endp

;------------------------------------------------------------------------
; strstr: Encuentra la primera ocurrencia de la subcadena buscada en una cadena
; 
; IN:
;     push offset cadena donde buscamos la subcadena (haystack)
;     push offset subcadena buscada (needle)
;     call STRSTR
; 
; OUT: 
;     AX - indice subcadena o -1 si no se encuentra
; 
; OJO!!
;   1. Asume que haystack esta en ES y needle en DS
;   2. Asume cadenas terminadas en \x00
;------------------------------------------------------------------------
strstr  proc    near
    push bp
    mov bp,sp
    push bx,cx,dx,si,di ;OJO! no optimizar con pusha/popa (ver fin rutina)
 
; ahora la pila es
; bp+0 => old BP
; bp+2 => @ret
; bp+4 => subcadena buscada
; bp+6 => cadena donde buscamos
; registros pusheados
 
    _cadena_grande   equ [word bp+6]
    _buscada         equ [word bp+4]

    xor cx,cx                   ;contador
    mov si,_cadena_grande
    mov bx,_buscada
COMPROBAR:   
    mov dl,byte es:[si]
    test dl,dl                  ;¿hemos llegado al final de la cadena?
    je NO_ENCONTRADA            ;si, no match
    cmp dl,byte [bx]            ;¿cadena_grande[cx] == buscada[0]?
    je ENCONTRADA1
SIGUIENTE:
    inc cx
    inc si
    jmp short COMPROBAR
ENCONTRADA1:
    push cx,si,bx,di,dx
COMPROBAR_SIGUIENTE_CHAR:
; ya sabemos que el primer caracter coincide
    inc si                      ;indice_cadena_grande (g++)
    inc bx                      ;indice_cadena_buscada (b++)
    mov dl,byte es:[si]         ;grande[g]
    mov dh,byte [bx]            ;buscada[b]
;; if (grande[g] == \x00 && buscada[b] == \x00) then match else break
    test dl,dl
    jne CONTINUAR
    test dh,dh
    je ENCONTRADA
;;
; grande==\x00 && buscada!=\x00
    pop dx,di,bx,si,cx
    jmp short SIGUIENTE
CONTINUAR:
; if (buscada[b] == \x00) then match
    test dh,dh
    je ENCONTRADA
; if (haystack[h] == needle[n])
    cmp dl,dh
    je COMPROBAR_SIGUIENTE_CHAR     ;match, comprobar siguiente caracter
; else (no match) break
    pop dx,di,bx,si,cx
    jmp short SIGUIENTE
NO_ENCONTRADA:
    mov ax,-1
    jmp short END_STRSTR
ENCONTRADA:
    pop dx,di,bx,si,cx
    xchg ax,cx
END_STRSTR:
    pop di,si,dx,cx,bx
    mov sp,bp
    pop bp
    ret 4               ;2*(2 parametros)
strstr  endp

indicecadena proc near
; IN:
;  push segment cadena_juego
;  push offset cadena_juego
;  push segment cadenas_tsr_en (array, ej: verbos_tsr_en, nombres_tsr_en)
;  push offset cadenas_tsr_en (array, ej: verbos_tsr_en, nombres_tsr_en)
;  push num. elementos array
;  call indicecadena

; OUT: AX=NUM_INDICE SI SE ENCUENTRA CADENA
; OUT: AX=-1 NO SE ENCUENTRA CADENA

    push bp
    mov bp,sp
    push cx,bx      ;backup

; ahora la pila es
; bp+0  => old BP
; bp+2  => @ret
; bp+4  => num. cadenas_tsr
; bp+6  => offset cadenas_tsr_en (array)
; bp+8  => segment cadenas_tsr_en (array)
; bp+0a => offset cadena_juego
; bp+0c => segment cadena_juego

    _segment_cadena_juego   equ [word bp+0c]
    _offset_cadena_juego    equ [word bp+0a]
    _segment_cadena_tsr     equ [word bp+8]     ;array
    _offset_cadena_tsr      equ [word bp+6]     ;array
    _num_cadenas_tsr        equ [word bp+4]

    mov cx,_num_cadenas_tsr
    mov bx,_offset_cadena_tsr   ;array
    dec bx
    dec bx
@@loop_buscacadena: inc bx
    inc bx                      ;siguiente cadena_tsr_en
    push _segment_cadena_juego
    push _offset_cadena_juego
    push _segment_cadena_tsr
    push Word Ptr ds:[bx]       ;offset cadena_tsr_en
    call strcmp
    test ax,ax                  ;¿match?
loopnz @@loop_buscacadena
    test ax,ax                  ;la ultima cadena podria ser un match
    jnz @@end_indicecadena      ;no match, bye
    xchg ax,bx                  ;match
    sub ax,_offset_cadena_tsr   ;array
    shr ax,1
@@end_indicecadena: pop bx,cx   ;restore
    mov sp,bp
    pop bp
    ret 0ah                     ;2*(5 parametros)
indicecadena endp
