; This file is part of PALESTRA
;
; PALESTRA v0.1, abr-2021, GPL-2.0-or-later, vlan7 <https://www.vlan7.org>
; un TSR para traducir al castellano la aventura grafica/conversacional
; Les Manley in: Search for the King v2.0
;
; PALESTRA is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 2 of the License, or
; (at your option) any later version.
;
; PALESTRA is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with PALESTRA.  If not, see <https://www.gnu.org/licenses/>.
;
; why PALESTRA?
; $ grep les /usr/share/dict/spanish |grep tra
; palestra

; ********************************************************************
; *                                                                  *
; *     R U T I N A S     S T R I N G . H     P A R A     I N T s    *
; *                                                                  *
; ********************************************************************

strcmp_palabra proc near
;IN: offset palabra_tsr (palabra=verbo|nombre)
;
;OUT: AX=longitud palabra (>0) si se encuentra, 0 si no se encuentra
;
;NOTA: Asume que palabra_tsr esta en ds y hay que compararla con es:di
;
    push bp
    mov bp,sp
    push    si,di,cx   ;backup

;ahora la pila es:
;bp+0 => old BP
;bp+2 => @ret
;bp+4 => offset palabra_tsr
;registros pusheados

    mov si,word [bp+4]  ;palabra_tsr
    xor cx,cx           ;longitud palabra
;    cld
@@loop_strcmp_palabra:  lodsb
    test al,al          ;¿palabra_es[i] == 0?
    jz ESPACIO_O_NULL6  ;SI: comprobamos que despues de palabra introducida
                        ; hay un espacio o un \x00
    cmp al,byte es:[di] ;NO: comparamos
    jne @@nomatch_strcmp_palabra
    inc cx              ;match
    inc di              ;siguiente caracter
jmp short @@loop_strcmp_palabra
ESPACIO_O_NULL6:    cmp byte es:[di],0  ;null
    jz @@exit_strcmp_palabra
    cmp byte es:[di],20h                ;espacio
    jz @@exit_strcmp_palabra
    cmp byte es:[di],3fh                ;?
    jz @@exit_strcmp_palabra
    cmp byte es:[di],2eh                ;.
    jz @@exit_strcmp_palabra
    cmp byte es:[di],2ch                ;,
    jz @@exit_strcmp_palabra
@@nomatch_strcmp_palabra:   xor cx,cx   ;longitud palabra
@@exit_strcmp_palabra:  xchg ax,cx      ;AX=longitud palabra
    pop cx,di,si
    mov sp,bp
    pop bp
    ret 2                               ;2*(1 param)
strcmp_palabra endp

strcmp_palabras proc near
;;comparar todas las palabra_en_tsr (ej: verbo_en_tsr, nombre_en_tsr) con es:di
;
;IN: offset palabra_en_tsr (palabra=verbo|nombre)
;
;OUT: AX=indice palabra_en_tsr (>0) si se encuentra, 0 si no se encuentra

    push bp
    mov bp,sp
    push    bx,cx,dx   ;backup

;ahora la pila es:
;bp+0 => old BP
;bp+2 => @ret
;bp+4 => num. palabras (ej: _verbo_array_cx, _nombre_array_cx)
;bp+6 => offset array palabra_en_tsr (ej: verbo_en_array, nombre_en_array)
;registros pusheados

    _palabra_en_tsr  equ [word bp+6]    ;palabra_en_tsr (array)

    mov bx,word _palabra_en_tsr
    mov cx,word [bp+4]                  ;num. palabras
    dec bx
    dec bx
@@loop_strcmp_palabras: inc bx
    inc bx                              ;siguiente palabra_en_tsr
    push Word Ptr [bx]                  ;palabra_en_tsr
    call strcmp_palabra
    test ax,ax                          ;¿match?
loopz @@loop_strcmp_palabras
    test ax,ax                          ;la ultima cadena podria ser un match
    jz @@exit_strcmp_palabras           ;no match, bye
    add di,ax                           ;match
;a veces ultimo verbo|nombre introducido no coincide con verbo|nombre desc_juego
;ej: place desk on chair (comando introducido)
;    You can't take the desk. (desc_juego)
;              ^-es:di
;    que traduzca "No puedes coger escritorio."
    mov ax,bx
    mov bx,Word _palabra_en_tsr         ;array
    sub ax,bx
;    shl ax,1        
@@exit_strcmp_palabras:
    pop dx,cx,bx                        ;restore
    mov sp,bp
    pop bp
    ret 4                               ;2*(2 params)
strcmp_palabras endp

strcmp_plus proc near
;------------------------------------------------------------------------
; compara s1 con s2 (cadenas).
; 
; in:
;   push segment s1     ;es ;desc_juego
;   push offset s1      :di ;desc_juego
;   push segment s2     ;ds ;desc_tsr_en
;   push offset s2      ;si ;desc_tsr_en
;   call strcmp_plus
;
; out:
;   ax>0 match
;   ax=0 no match
;
; nota: si se encuentra un %v y procede, actualiza ultimo_verbo_bx
;------------------------------------------------------------------------

    push bp
    mov bp,sp
    push    es,di,ds,si,bx,cx   ;backup

;ahora la pila es:
;bp+0 => old BP
;bp+2 => @ret
;bp+4 => offset s2      ;desc_tsr_en
;bp+6 => segment s2     ;desc_tsr_en
;bp+8 => offset s1      ;desc_juego
;bp+0ah => segment s1   ;desc_juego
;registros pusheados

    mov es,word [bp+0ah]    ;desc_juego
    mov di,word [bp+8]      ;desc_juego
    mov ds,word [bp+6]      ;desc_tsr_en
    mov si,word [bp+4]      ;desc_tsr_en
    xor cx,cx
    cld
@@loop_strcmp_plus: cmp Byte Ptr ds:[si],0  ;¿desc_tsr_en recorrida entera?
    jz @@fin_strcmp_plus        ;match, por ahora
    cmp byte ds:[si],2ah        ;¿desc_tsr_en == '*'?
    je @@es_asterisco
;    cmp word ds:[si],6e25h     ;¿desc_tsr_en == '%n'?
;    je @@es_nombre
    cmp word ds:[si],7625h      ;¿desc_tsr_en == '%v'?
    je @@es_verbo
    cmpsb
    jne @@nomatch_strcmp_plus   ;no match
    inc cx                      ;match, por ahora
jmp short @@loop_strcmp_plus    ;match, por ahora. repetir

@@es_verbo:
;; si ultimo_verbo_en_introducido match con verbo_desc_juego -> usamos ultimo_verbo_en_introducido que ya tenemos
;; si no match -> recorremos todos los verbo_en_tsr del diccionario, y si alguno match, actualizamos ultimo_verbo_en_introducido con el verbo que hace match
    inc si
    inc si                  ;saltar '%v'
    push    ds              ;backup
    push cs
    pop ds
    lea bx,verbo_en_array
    add bx,ultimo_verbo_bx
    push Word Ptr [bx]      ;offset ultimo_verbo_en_introducido
    call strcmp_palabra
    test ax,ax              ;¿match?
    jnz @@fin_es_verbo
    push offset verbo_en_array  ;no match, comparar verbo_en_juego con todos los verbos_en_tsr
    push _verbo_array_cx
    call strcmp_palabras
    test ax,ax              ;¿match?
    jz @@fin_es_verbo       ;no match
    mov ultimo_verbo_bx,ax  ;match, update ultimo_verbo_bx
    pop ds                  ;restore
jmp short @@loop_strcmp_plus
@@fin_es_verbo: lea bx,verbo_en_array
    add bx,ultimo_verbo_bx
    push Word Ptr [bx]      ;offset ultimo_verbo_en_introducido
    call strlen
    add di,ax
    pop ds                  ;restore
jmp short @@loop_strcmp_plus

;@@es_nombre:
;    inc si
;    inc si                  ;saltar '%n'
;    push    ds              ;backup
;    push cs
;    pop ds
;    lea bx,nombre_en_array
;    add bx,ultimo_nombre_bx
;    push Word Ptr [bx]      ;offset ultimo_nombre_en_introducido
;    call strcmp_palabra
;    test ax,ax              ;¿match?
;    jnz @@fin_es_nombre
;    push offset nombre_en_array  ;no match, comparar nombre_en_juego con todos los nombres_en_tsr
;    push _nombre_array_cx
;    call strcmp_palabras
;    test ax,ax              ;¿match?
;    jz @@fin_es_nombre       ;no match
;    mov ultimo_nombre_bx,ax  ;match, update ultimo_nombre_bx
;    pop ds                  ;restore
;jmp short @@loop_strcmp_plus
;@@fin_es_nombre:
;    lea bx,nombre_en_array
;    add bx,ultimo_nombre_bx
;    push Word Ptr [bx]      ;offset ultimo_nombre_en_introducido
;    call strlen
;    add di,ax
;    pop ds                  ;restore
;jmp short @@loop_strcmp_plus

@@es_asterisco:
;;IMPORTANTE: asumimos que despues del * no hay ningun %loquesea (ni otro *).
;;(si algo falla: revisar descripciones en EMSDATA.ASM y UMBDATA.ASM)
    inc si      ;saltar '*'
;The * is already open.
;     ^--ds:si
;The drawer is already open.
;    ^--es:di    
    push di
    push si
    call strstr
    mov cx,ax
    inc cx      ;si ax=-1, en @@exit_strcmp_plus ax=0 (no match)
                ;si no, en @@exit_strcmp_plus ax>0 (match, ax=longitud)
    jmp short @@exit_strcmp_plus
@@fin_strcmp_plus:
    cmp byte es:[di],0      ;¿ultimo caracter desc_juego == NULL?
    jz @@exit_strcmp_plus   ;match
    cmp byte es:[di],20h    ;¿espacio?
    je @@exit_strcmp_plus   ;match
@@nomatch_strcmp_plus:  xor cx,cx   ;no match
@@exit_strcmp_plus:
    mov ax,cx               ;longitud. ej:'drawer'+1(por el espacio que sigue)=7
    pop cx,bx,si,ds,di,es   ;restore
    mov sp,bp
    pop bp
    ret 8                   ;2*(4 params)
strcmp_plus endp

strcpy_plus proc near
;------------------------------------------------------------------------
; copia s2 (origen) en s1 (destino)
; 
; in:
;   push segment s1 (destino)   ;es ;desc_juego
;   push offset s1 (destino)    ;di ;desc_juego
;   push segment s2 (origen)    ;ds ;desc_tsr_es
;   push offset s2 (origen)     ;si ;desc_tsr_es
;   call strcmp_plus
;
; out:
;   ax>0 match
;   ax=0 no match
;
; OJO!!
;   1. Asume cadenas terminadas en \x00
;   2. Asume s1 (destino) suficientemente grande
;------------------------------------------------------------------------
    push bp
    mov bp,sp
    push    bx,cx,si,di,es,ds   ;backup
 
;ahora la pila es:
;bp+0 => old BP
;bp+2 => @ret
;bp+4 => offset s2 (origen)     ;si ;desc_tsr_es
;bp+6 => segment s2 (origen)    ;ds ;desc_tsr_es
;bp+8 => offset s1 (destino)    ;di ;desc_juego
;bp+0ah => segment s1 (destino) ;es ;desc_juego
;registros pusheados

    mov es,word [bp+0ah]    ;desc_juego
    mov di,word [bp+8]      ;desc_juego
    mov ds,word [bp+6]      ;desc_tsr_es
    mov si,word [bp+4]      ;desc_tsr_es

@@loop_strcpy_plus: cmp Byte Ptr ds:[si],0      ;¿desc_tsr_es copiada entera?
;    jz @@fin_strcpy_plus                       ;bye
jnz CHAPUZABLABLA
jmp @@fin_strcpy_plus                        ;bye
CHAPUZABLABLA:    cmp Word Ptr ds:[si], 7625h                 ;¿desc_tsr_es == '%v'?
    je @@strcpy_ultimo_verbo
    cmp Word Ptr ds:[si],6e25h                  ;¿desc_tsr_es == '%n'?
    je @@strcpy_ultimo_nombre
    cmp Word Ptr ds:[si], 3225h                 ;¿desc_tsr_es == '%2'?
    je @@strcpy_ultimo_nombre2
    movsb
jmp short @@loop_strcpy_plus

@@strcpy_ultimo_verbo:
    push    ds  ;backup
    push cs
    pop ds
;¿%v = " "?
;si, write to memory "No puedes hacer eso." (definir cadena en Datos Residentes)
    cmp ultimo_verbo_bx,0       ;¿parser guardó indice ultimo verbo introducido?
    jge @@parser_sabe_verbo     ;si, usamos ese valor
    mov di,word [bp+8]
    mov si,offset safemsg       ;"No puedes hacer eso."
    mov cx,21d
    rep movsb                   ;write to mem (desc_juego)
    jmp short @@fin_strcpy_plus
@@parser_sabe_verbo:
;; vamos a escribir ultimo verbo_es introducido en descripcion que daria el juego
    lea bx,verbo_es_array
    add bx,ultimo_verbo_bx
    call CHAPUZA
    pop ds                      ;restore
jmp short @@loop_strcpy_plus

@@strcpy_ultimo_nombre:
    push    ds                  ;backup
    push cs
    pop ds
    cmp ultimo_nombre_bx,0      ;¿parser guardó indice ultimo nombre introducido?
    jge @@parser_sabe_nombre    ;si, usamos ese valor
    mov di,word [bp+8]
    mov si,offset safemsg       ;"No puedes hacer eso."
    mov cx,21d
    rep movsb                   ;write to mem (desc_juego)
    jmp short @@fin_strcpy_plus
@@parser_sabe_nombre:
;; vamos a escribir ultimo nombre_es introducido en descripcion que daria el juego
    lea bx,nombre_es_array
    add bx,ultimo_nombre_bx
    call CHAPUZA
    pop ds                      ;restore
jmp short @@loop_strcpy_plus

@@strcpy_ultimo_nombre2:
    push    ds                  ;backup
    push cs
    pop ds
    cmp ultimo_nombre2_bx,0     ;¿parser guardó indice ultimo nombre2 introducido?
    jge @@parser_sabe_nombre2   ;si, usamos ese valor
    mov di,word [bp+8]
    mov si,offset safemsg       ;"No puedes hacer eso."
    mov cx,21d
    rep movsb                   ;write to mem (desc_juego)
    jmp short @@fin_strcpy_plus
@@parser_sabe_nombre2:
;; vamos a escribir ultimo nombre2_es introducido en descripcion que daria el juego
    lea bx,nombre_es_array
    add bx,ultimo_nombre2_bx
    call CHAPUZA
    pop ds                      ;restore
jmp short @@loop_strcpy_plus

@@fin_strcpy_plus:
    mov byte es:[di],0          ;terminar desc_juego
    pop ds,es,di,si,cx,bx       ;restore
    mov sp,bp
    pop bp
    ret 8                       ;2*(4 params)
strcpy_plus endp

CHAPUZA PROC NEAR
        inc si
        inc si  ; saltar '%x'       ; saltamos %loquesea
        push si
        mov si, Word Ptr [bx]
        push si
        call STRLEN
        mov cx, ax
        rep movsb       ; copiamos ultima palabra_es introducida en ES:DI
        pop si          ; volvemos a apuntar DS:SI a la desc_es[i]
        ret
CHAPUZA ENDP
